@@ -56,6 +56,7 @@ static EventField fields[] = {
   { "message-token", G_TYPE_STRING, "Headers.value" },
   /* FIXME: these should really be in plugins */
   { "channel", G_TYPE_STRING, "Events.channel" },
+  { "outgoing", G_TYPE_BOOLEAN, "Events.outgoing" },
   { "free-text", G_TYPE_STRING, "Events.free_text" },
   { NULL, 0, NULL }
 };
@@ -154,6 +155,7 @@ static const gchar *db_schema_sql[] = {
     "start_time INTEGER NOT NULL," \
     "end_time INTEGER," \
     "is_read INTEGER DEFAULT 0," \
+    "outgoing BOOL DEFAULT 0," \
     "flags INTEGER DEFAULT 0," \
     "bytes_sent INTEGER DEFAULT 0," \
     "bytes_received INTEGER DEFAULT 0," \
@@ -246,12 +248,6 @@ rtcom_el_db_schema_get_sql ()
   return db_schema_sql;
 }
 
-gint
-rtcom_el_db_schema_get_n_items ()
-{
-  return sizeof(db_schema_sql) / sizeof (const gchar *) - 1;
-}
-
 void
 rtcom_el_db_schema_get_mappings (const gchar **out_selection,
     GHashTable **out_mapping, GHashTable **out_typing)
@@ -335,6 +331,270 @@ _handle_corrupted (const gchar *fname, gboolean try_repairing,
     return NULL;
 }
 
+static gboolean
+_internal_convert_v0 (rtcom_el_db_t db)
+{
+  gint plugin_id = 0;
+  gint service_id = 0;
+  int i;
+  const gchar *services[] = { "CHAT", "CALL", "SMS", NULL };
+  struct {
+      const gchar *name;
+      const gchar *desc;
+  } eventtypes[] = {
+      { "RTCOM_EL_EVENTTYPE_CHAT_NOTICE", "Notice" },
+      { "RTCOM_EL_EVENTTYPE_CHAT_ACTION", "Action message" },
+      { "RTCOM_EL_EVENTTYPE_CHAT_AUTOREPLY", "Autoreply message" },
+      { "RTCOM_EL_EVENTTYPE_CHAT_JOIN", "Group chat joined" },
+      { "RTCOM_EL_EVENTTYPE_CHAT_LEAVE", "Group chat leave" },
+      { "RTCOM_EL_EVENTTYPE_CHAT_TOPIC", "Group chat topic change" },
+      { NULL, NULL }
+  };
+  struct {
+      const gchar *name;
+      gint value;
+      const gchar *desc;
+  } flags[] = {
+      { "RTCOM_EL_FLAG_CHAT_GROUP", 1, "Groupchat message" },
+      { "RTCOM_EL_FLAG_CHAT_ROOM", 2, "Groupchat is room with channel_id" },
+      { "RTCOM_EL_FLAG_CHAT_OPAQUE", 4, "Channel identifier is opaque" },
+      { "RTCOM_EL_FLAG_CHAT_OFFLINE", 8, "Offline message" },
+      { NULL, 0, NULL }
+  };
+
+#define runsql(x) rtcom_el_db_exec (db, NULL, NULL, (x), NULL)
+#define vrunsql(x...) rtcom_el_db_exec_printf (db, NULL, NULL, NULL, x)
+
+  /* This is done in transaction started by the caller, but we will
+   * commit/rollback the transaction ourselves at the end. */
+
+  /* Add two new columns to Events table. */
+  if (!runsql ("ALTER TABLE Events ADD COLUMN outgoing BOOL DEFAULT 0;"))
+    goto err;
+
+  if (!runsql ("ALTER TABLE Events ADD COLUMN mc_profile BOOL DEFAULT 0;"))
+    goto err;
+
+  /* Set outgoing column for any outbound events. */
+  if (!runsql ("UPDATE Events SET outgoing = 1 WHERE event_type_id IN "
+      "(SELECT id FROM EventTypes WHERE name LIKE '%_OUTBOUND');"))
+    goto err;
+
+  /* For each recognised service, map all events to a single event
+   * type (*_INBOUND), rename the event type to the new name and
+   * delete the other now obsolete, event type (_OUTBOUND). */
+  for (i = 0; services[i] != NULL; i++)
+    {
+      gint id = 0;
+      gint old_id = 0;
+
+      if (!rtcom_el_db_exec_printf (db, rtcom_el_db_single_int, &id, NULL,
+        "SELECT id FROM EventTypes WHERE name = "
+          "'RTCOM_EL_EVENTTYPE_%s_INBOUND'", services[i]))
+        goto err;
+
+      if (!id)
+        goto err;
+
+      if (!rtcom_el_db_exec_printf (db, rtcom_el_db_single_int, &old_id, NULL,
+        "SELECT id FROM EventTypes WHERE name = "
+          "'RTCOM_EL_EVENTTYPE_%s_OUTBOUND'", services[i]))
+        goto err;
+
+      if (!old_id)
+        goto err;
+
+      if (!vrunsql ("UPDATE Events SET event_type_id = %d WHERE "
+          "event_type_id = %d", id, old_id))
+        goto err;
+
+      if (!vrunsql ("UPDATE EventTypes SET name = 'RTCOM_EL_EVENTTYPE_%s%s' "
+          "WHERE name = 'RTCOM_EL_EVENTTYPE_%s_INBOUND'", services[i],
+          (g_strcmp0 (services[i], "CALL")) ? "_MESSAGE" : "", services[i]))
+        goto err;
+
+      if (!vrunsql ("DELETE FROM EventTypes WHERE name = "
+        "'RTCOM_EL_EVENTTYPE_%s_OUTBOUND'", services[i]));
+    }
+
+  if (!rtcom_el_db_exec (db, rtcom_el_db_single_int, &plugin_id,
+      "SELECT plugin_id FROM EventTypes WHERE name = "
+          "'RTCOM_EL_EVENTTYPE_CHAT_MESSAGE' LIMIT 1", NULL))
+    goto err;
+
+  if (!plugin_id)
+    goto err;
+
+  /* Add new CHAT event types */
+  for (i = 0; eventtypes[i].name != NULL; i++)
+    {
+      if (!vrunsql ("INSERT INTO EventTypes (name, plugin_id, desc) VALUES "
+          "(%Q, %d, %Q)", eventtypes[i].name, plugin_id, eventtypes[i].desc))
+        goto err;
+    }
+
+  if (!rtcom_el_db_exec_printf (db, rtcom_el_db_single_int, &service_id, NULL,
+      "SELECT id FROM Services WHERE plugin_id = %d", plugin_id))
+    goto err;
+
+  if (!service_id)
+    goto err;
+
+  /* Add new CHAT flags */
+  for (i = 0; flags[i].name != NULL; i++)
+    {
+      if (!vrunsql ("INSERT INTO Flags (service_id, name, value, desc) VALUES "
+          "(%d, %Q, %d, %Q)", service_id, flags[i].name, flags[i].value,
+          flags[i].desc))
+        goto err;
+    }
+
+  /* Now, if we've updated S3 db, we have to clean up some old triggers; this
+   * is copy/paste from the db initialisation. */
+  
+  runsql ("DROP TRIGGER IF EXISTS gc_update_ev_add2;");
+  runsql ("DROP TRIGGER IF EXISTS gc_update_ev_add3;");
+  runsql ("DROP TRIGGER IF EXISTS gc_update_ev_update;");
+
+  runsql ("CREATE TRIGGER IF NOT EXISTS gc_update_ev_add4 AFTER INSERT ON Events " \
+       "FOR EACH ROW WHEN NEW.group_uid IS NOT NULL BEGIN " \
+           "UPDATE GroupCache SET event_id = NEW.id, total_events = total_events + 1, " \
+           "read_events = read_events + NEW.is_read, flags = flags | NEW.flags " \
+           "WHERE group_uid = NEW.group_uid; " \
+       "END;");
+
+  runsql ("CREATE TRIGGER IF NOT EXISTS gc_update_ev_update AFTER UPDATE ON Events " \
+       "FOR EACH ROW WHEN NEW.group_uid IS NOT NULL BEGIN " \
+           "UPDATE GroupCache SET " \
+               "read_events = read_events - OLD.is_read + NEW.is_read, "\
+               "flags = (flags & (~OLD.flags)) | NEW.flags " \
+               "WHERE group_uid = NEW.group_uid; " \
+        "END;");
+
+  /* We have already set user_version to 1, but the backup process
+   * overwrote that, so we have to do it again. */
+  if (!runsql ("PRAGMA user_version = 1"))
+    goto err;
+
+  if (!rtcom_el_db_commit (db, NULL))
+    goto err;
+
+  return TRUE;
+
+err:
+  g_warning("%s: sqlite error: %s", G_STRFUNC,
+      sqlite3_errmsg (db));
+  rtcom_el_db_rollback (db, NULL);
+  return FALSE;
+
+#undef runsql
+#undef vrunsql
+}
+
+gboolean
+rtcom_el_db_convert_from_db0 (const gchar *fname, const gchar *old_fname)
+{
+  rtcom_el_db_t db = NULL;
+  rtcom_el_db_t old_db = NULL;
+  sqlite3_backup *bkp;
+  gint ret;
+  int cnt;
+  gboolean success;
+
+  /* To speed things up in most likely case (new database already exists),
+   * we first check whether the new database file already exists. */
+  if (g_access (fname, 0) == 0)
+      return TRUE;
+
+  /* We don't want to start converting into a new database if we can't
+   * reach the old one. */
+  if (g_access (old_fname, 0) == -1)
+      return FALSE;
+
+  ret = sqlite3_open (old_fname, &old_db);
+
+  if (ret != SQLITE_OK)
+      return FALSE;
+
+  ret = sqlite3_open (fname, &db);
+  if (ret != SQLITE_OK)
+    {
+      sqlite3_close (old_db);
+      return FALSE;
+    }
+
+  /* We'd rather be a bit slower with on-disk journal then screw up db
+   * containing user data. */
+  rtcom_el_db_exec (db, NULL, NULL, "PRAGMA journal_mode = DELETE;", NULL);
+
+  if (!rtcom_el_db_exec (db, rtcom_el_db_single_int, &ret,
+      "PRAGMA user_version", NULL))
+    {
+      sqlite3_close (old_db);
+      sqlite3_close (db);
+      return FALSE;
+    }
+
+  /* It's already been converted. */
+  if (ret == 1)
+    {
+      sqlite3_close (old_db);
+      sqlite3_close (db);
+      return TRUE;
+    }
+
+  /* While we're busy converting the database, another app might notice
+   * the file exists and try to open it. If user_version is not set
+   * immediately, it would fail miserably. */
+  rtcom_el_db_exec (db, NULL, NULL, "PRAGMA user_version = 1;", NULL);
+
+  if (!rtcom_el_db_transaction (db, TRUE, NULL))
+    {
+      sqlite3_close (old_db);
+      sqlite3_close (db);
+      return FALSE;
+    }
+
+  /* Start backing up old database into a new one (with the same format). */
+  bkp = sqlite3_backup_init (db, "main", old_db, "main");
+  if (!bkp)
+    {
+      sqlite3_close (old_db);
+      sqlite3_close (db);
+      return FALSE;
+    }
+
+  /* Try doing the backup, but don't retry forever if a source database is
+   * locked. */
+  for (cnt = 0; cnt < 100; cnt++)
+    {
+      ret = sqlite3_backup_step (bkp, -1);
+
+      if ((ret != SQLITE_BUSY) && (ret != SQLITE_LOCKED))
+          break;
+
+      g_usleep (G_USEC_PER_SEC / 100);
+    }
+
+  ret = sqlite3_backup_finish (bkp);
+
+  if (ret != SQLITE_OK)
+    {
+      sqlite3_close (old_db);
+      sqlite3_close (db);
+      return FALSE;
+    }
+
+  sqlite3_close (old_db);
+
+  /* Now we can do the conversion. */
+  success = _internal_convert_v0 (db);
+
+  sqlite3_close (db);
+  return success;
+}
+
+
 /* Opens a new SQLite3 database, creating and initialising it if
  * neccessary. If the existing database is corrupted, it's deleted
  * and a new database is created.
@@ -373,28 +633,34 @@ _internal_open (const gchar *fname, gboolean try_repairing)
   sqlite3_profile (db, profile_cb, NULL);
 #endif
 
-  /* If we can't even get the user version, something's horribly wrong. */
-  if (!rtcom_el_db_exec (db, rtcom_el_db_single_int, &user_version,
-      "PRAGMA user_version;", &err))
+  /* Quick check to see if the database is valid. */
+  if (!rtcom_el_db_exec (db, NULL, NULL, "PRAGMA quick_check;", &err))
     {
-      g_warning ("%s: can't get schema version", G_STRFUNC);
       sqlite3_close (db);
 
       if (err->code == RTCOM_EL_DATABASE_CORRUPTED)
         {
           return _handle_corrupted (fname, try_repairing, err);
         }
-
-      g_error_free (err);
-      return NULL;
+      else
+        {
+          g_error_free (err);
+          return NULL;
+        }
     }
 
+  rtcom_el_db_exec (db, rtcom_el_db_single_int, &user_version,
+      "PRAGMA user_version;", NULL);
+
   /* If schema hasn't been defined, we can attempt to do so. Race condition
      here is mostly harmless because CREATEs are guarded by IF NOT EXIST. But,
      we're doing the detection to avoid slow startup and messing with other
      app' queries that might be happening at the time. */
   if (user_version < REQUIRED_USER_VERSION)
     {
+      /* If schema hasn't been defined, we can attempt to do so. Race condition
+       * here is harmless because CREATEs are guarded by IF NOT EXIST. But,
+       * we're doing the detection to avoid slow startup. */
       gint i;
 
       g_chmod (fname, S_IRUSR | S_IWUSR);
@@ -574,13 +840,35 @@ rtcom_el_db_exec (rtcom_el_db_t db, GFunc cb, gpointer user_data,
   g_assert (sql);
 
   ret = sqlite3_prepare_v2 (db, sql, -1, &stmt, NULL);
-  if (ret != SQLITE_OK)
-    {
-      g_warning ("%s: can't compile SQL statement \"%s\": %s", G_STRFUNC, sql,
-         sqlite3_errmsg (db));
-      g_set_error (error, RTCOM_EL_ERROR, RTCOM_EL_INTERNAL_ERROR,
-         "Can't compile SQL statement: %s", sqlite3_errmsg (db));
-      return FALSE;
+  switch (ret)
+  {
+      case SQLITE_DONE:
+      case SQLITE_ROW:
+      case SQLITE_OK:
+          /* All is well. */
+          break;
+
+      case SQLITE_FULL:
+      case SQLITE_IOERR:
+          g_debug ("%s: database full or I/O error", G_STRFUNC);
+          g_set_error (error, RTCOM_EL_ERROR, RTCOM_EL_DATABASE_FULL,
+             "Database full");
+          return FALSE;
+
+      case SQLITE_CORRUPT:
+      case SQLITE_FORMAT:
+      case SQLITE_NOTADB:
+          g_debug ("%s: database corrupted", G_STRFUNC);
+          g_set_error (error, RTCOM_EL_ERROR, RTCOM_EL_DATABASE_CORRUPTED,
+             "Database corrupted");
+          return FALSE;
+
+      default:
+          g_warning ("%s: can't compile SQL statement \"%s\": %s", G_STRFUNC, sql,
+             sqlite3_errmsg (db));
+          g_set_error (error, RTCOM_EL_ERROR, RTCOM_EL_INTERNAL_ERROR,
+             "Can't compile SQL statement: %s", sqlite3_errmsg (db));
+          return FALSE;
     }
 
   do
